# -*- coding: utf-8 -*-
# @Time    : 2019/10/24 17:50
# @E-Mail  : aberstone.hk@gmail.com
# @File    : __version__.py
# @Software: PyCharm

# __title__ = 'bloomfilter'
__title__ = 'iron_man'
__description__ = 'C & Python Bloomfilter Based Memeroy || Redis .'
__version__ = '1.0.2'
__build__ = 0x022200
__author__ = 'Aberstone'
__author_email__ = 'aberstone.hk@gmail.com'
__license__ = 'MIT License'
__copyright__ = 'Copyright 2019 Aberstone'
__cake__ = u'\u2728 \U0001f370 \u2728'
__url__ = "http://gitlab.nuist.pro:3000/aberstone/IronMan.git"


# -*- coding: utf-8 -*-
# @Time    : 2019-05-29 10:35
# @E-Mail  : wujifan1106@gmail.com
# @Site    : 
# @File    : bloomfilter.py
# @Software: PyCharm
import math
import traceback
from typing import AnyStr

from CBloomfilter import CBloomfilter


class LocalBloomFilter():

    def __init__(self, capacity, error, prime_length=True):
        self.bf = CBloomfilter(capacity, error, prime_length)
        self.bitmap = bytes(int(self.bf.bits / 8) + 1)

    def add(self, data):
        if isinstance(data, (list, tuple)):
            for v in data:
                assert isinstance(v, str), 'add() arg must be a str or list/tuple of strings'
                self.bf.add(self.bitmap, v)
        else:
            assert isinstance(data, str), 'add() arg must be a str or list/tuple of strings'
            self.bf.add(self.bitmap, data)

    def is_contain(self, data):
        if isinstance(data, (list, tuple)):
            for v in data:
                assert isinstance(v, str), 'is_contain() arg must be a str or list/tuple of strings'
            return [self.bf.is_contain(self.bitmap, v) for v in data]
        else:
            assert isinstance(data, str), 'is_contain() arg must be a str or list/tuple of strings'
            return self.bf.is_contain(self.bitmap, data)

    def clean(self):
        self.bf.clean_bitmap(self.bitmap)


class RedisBloomFilter():
    def __init__(self, capacity, error, redis_conn, prime_length=True, filter_prefix="BloomFilter"):
        self.bf = CBloomfilter(capacity, error, prime_length)
        total_memory_by_mb = math.ceil(self.bf.bits / 8 / 1024 / 1024)
        self.mem_block_counts = math.ceil(total_memory_by_mb / 512)
        self.filter_prefix = filter_prefix
        self.redis_conn = redis_conn

    def _add(self, data: AnyStr):
        assert isinstance(data, str), 'add() arg must be a str or list/tuple of strings'
        offset = self.bf.hash(data)
        key = self.filter_prefix + "_" + self._cal_data_block(data)
        with self.redis_conn.pipeline() as pipe:
            for o in offset:
                pipe.setbit(key, o, 1)
            pipe.execute()

    def _cal_data_block(self, data: AnyStr):
        return str(ord(data[0]) % self.mem_block_counts)

    def add(self, data):
        if isinstance(data, (list, tuple)):
            for v in data:
                self._add(v)
        else:
            self._add(data)

    def delete(self, key, data):
        if isinstance(data, (list, tuple)):
            offset = []
            for v in data:
                assert isinstance(v, str), 'add() arg must be a str or list/tuple of strings'
                offset += self.bf.hash(v)
            with self.redis_conn.pipeline() as pipe:
                for o in offset:
                    pipe.setbit(key, o, 0)
                pipe.execute()

        else:
            assert isinstance(data, str), 'add() arg must be a str or list/tuple of strings'
            offset = self.bf.hash(data)
            with self.redis_conn.pipeline() as pipe:
                for o in offset:
                    pipe.setbit(key, o, 0)
                pipe.execute()

    def is_contain(self, data):
        key = self.filter_prefix + "_" + self._cal_data_block(data)
        try:
            assert isinstance(data, str), 'is_contain() arg must be a str'
            offset = self.bf.hash(data)
            with self.redis_conn.pipeline() as pipe:
                for o in offset:
                    pipe.getbit(key, o)
                results = pipe.execute()
                if sum(results) == self.bf.hashes:
                    return True
                return False
        except Exception:
            print(traceback.format_exc())
            return None

    def clean(self):
        for i in range(self.mem_block_counts):
            self.redis_conn.delete(f"{self.filter_prefix}_{i}")


if __name__ == '__main__':
    f = RedisBloomFilter(100000000, 0.0001, None, True, "test")
    f.add("abc")



cdef extern from "c_bloomfilter.h":
    ctypedef unsigned long int   uint64_t
    ctypedef unsigned int   uint32_t

    ctypedef struct cbloomfilterctxt:
        uint32_t        capacity;
        uint32_t        hashes;
        uint64_t        bits;
        float           error;
        uint64_t       *seeds;

    void clean_bitmap(char *bitmap, uint32_t len)
    void free_bloomfilter(cbloomfilterctxt * ctxt)
    void init_bloomfilter(cbloomfilterctxt * ctxt, uint32_t capacity, float error,bint prime_length)
    void bf_add(cbloomfilterctxt* ctxt,char *bitmap, const char *data, uint32_t len);
    bint bf_is_contain(cbloomfilterctxt* ctxt,char *bitmap, const char *data, uint32_t len);
    uint64_t hash(const char* data, uint32_t len, uint64_t seed, uint64_t bits);

cdef class CBloomfilter(object):
    cdef cbloomfilterctxt context

    property bits:
        def __get__(self):
            return self.context.bits

    property hashes:
        def __get__(self):
            return self.context.hashes
    def __cinit__(self, capacity, error, prime_length = True):
        init_bloomfilter(&self.context,capacity,error,prime_length)

    def __dealloc__(self):
        free_bloomfilter(&self.context)

    def add(self,bitmap,data):
        key = data.encode()
        bf_add(&self.context, bitmap,key,len(key))

    def is_contain(self,bitmap,data):
        key = data.encode()
        return bf_is_contain(&self.context, bitmap,key,len(key))

    def clean_bitmap(self,bitmap):
        clean_bitmap(bitmap, len(bitmap))

    def hash(self,data):
        key = data.encode()
        offset = []
        for i in range(self.hashes):
            seed = self.context.seeds[i]
            offset.append(hash(key,len(key),seed,self.context.bits))
        return offset



# -*- coding: utf-8 -*-
# @Time    : 2019/10/24 17:50
# @E-Mail  : aberstone.hk@gmail.com
# @File    : __init__.py.py
# @Software: PyCharm
from .py_bloomfilter import RedisBloomFilter,LocalBloomFilter


# -*- coding: utf-8 -*-
# @Time    : 2019-05-29 10:42
# @E-Mail  : wujifan1106@gmail.com
# @Site    : 
# @File    : setup.py
# @Software: PyCharm
import os
# from distutils.core import setup
from setuptools import setup
# here = os.path.abspath(os.path.dirname(__file__))

ext_files = [os.path.join("iron_man", "c_bloomfilter.c")]

kwargs = {}

from Cython.Distutils import build_ext
from Cython.Distutils import Extension

print('Building from Cython')
ext_files.append(os.path.join("iron_man", "py_bloomfilter.pyx"))
# ext_files.append("./bloomfilter/py_bloomfilter.pyx")
kwargs['cmdclass'] = {'build_ext': build_ext}

ext_modules = [Extension("CBloomfilter", ext_files)]
# ext_modules = cythonize(ext_files[0])
about = {}
with open(os.path.join('iron_man', '__version__.py'), 'r', encoding='utf-8') as f:
    exec(f.read(), about)

from os import path as os_path

this_directory = os_path.abspath(os_path.dirname(__file__))


def read_file(filename):
    with open(os_path.join(this_directory, filename), encoding='utf-8') as f:
        long_description = f.read()
    return long_description


setup(
    name=about['__title__'],
    version=about['__version__'],
    url=about['__url__'],
    description=about['__description__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    license=about['__license__'],
    ext_modules=ext_modules,
    # py_modules=['bloomfilter.__init__'],
    packages=['iron_man'],
    package_dir={
        'iron_man': "iron_man"
    },
    package_data={
        '': ['LICENSE', 'NOTICE'],
        "iron_man": ["c_bloomfilter.h", "bloomfilter.c_bloomfilter.h"]
    },
    # include_package_data=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: C',
        'Programming Language :: Cython',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    long_description=read_file('README.md'),
    long_description_content_type="text/markdown",
    **kwargs
)
